try:
    import matplotlib.pyplot as plt
except ImportError:
    print("matplotlib couldn't be imported, please install it.")
    import sys
    sys.exit(1)

import numpy as np
from typing import Sequence
from ZOCallable.ZOCallable import ZOCallable, ZOZOCallable

def plot_ZOCallable(*func: tuple[ZOCallable], vectorized = True, save_path: str = "", labels: Sequence[str] = []):
    """
    Plot one or mulitple ZOCallable to visualize.
    
    Params:
    ----
    - func: ZOCallable, the functions to plot.
    - vectorized: bool, specify whether the functions are vectorized or not.
    - save_path: str = "", if specified, the plot isn't showed but saved

    The functions are plotted for x from 0 to 1, x = y is plotted in the background.
    """
    plt.figure()
    plt.plot([0, 1], [0, 1], c='#333', linestyle='--', label='')
    x = np.linspace(0, 1, 101)
    if not labels:
        lbls = ["" for _ in func]
    else:
        lbls = labels
    for f, lbl in zip(func, lbls):
        if vectorized:
            plt.plot(x, f(x), label=lbl)
        else:
            plt.plot(x, [f(t) for t in x], label=lbl)
    plt.grid()
    if labels:
        plt.legend()
    plt.xlim(-0.01, 1.01)
    plt.scatter([0, 1], [0, 1], marker='o', label='', c='k')
    if save_path:
        plt.savefig(save_path)
    else:
        plt.show()

def plot_ZOZOCallable(*func: ZOZOCallable, vectorized = True, save_path: str = "", labels: Sequence[str] = []):
    """
    Plot one or mulitple ZOZOCallable to visualize.
    
    Params:
    ----
    - func: ZOCallable, the functions to plot.
    - vectorized: bool, specify whether the functions are vectorized or not.
    - save_path: str = "", if specified, the plot isn't showed but saved


    The functions are plotted for x from 0 to 1, x = y is plotted in the background.
    As ZOZOCallables should range in [0, 1], the ylim is fixed to [0, 1].
    """
    plt.figure()
    plt.plot([0, 1], [0, 1], c='#333', linestyle='--')
    x = np.linspace(0, 1, 101)
    if not labels:
        lbls = ["" for _ in func]
    else:
        lbls = labels
    for f, lbl in zip(func, lbls):
        if vectorized:
            plt.plot(x, f(x), label=lbl)
        else:
            plt.plot(x, [f(t) for t in x], label=lbl)
    plt.grid()
    if labels:
        plt.legend()
    plt.xlim(-0.01, 1.01)
    plt.ylim(-0.01, 1.01)
    plt.scatter([0, 1], [0, 1], marker='o', c='k')
    if save_path:
        plt.savefig(save_path)
    else:
        plt.show()



"""
ZOCallable is a python library defining functions ranging from 0 to 1. They can be used as transitions or non-linear gradients for example.
ZO stands for "Zero One".

ZOCallable contains two classes:
- ZOCallable, they represent functions mapping floats of [0, 1] to other floats, and mapping 0 to 0 and 1 to 1.
- ZOZOCallable, they represent functions mapping floats of [0, 1] to [0, 1], and mapping 0 to 0, and 1 to 1.

For both cases, a function called `verify_ZO(ZO)Callable` can be used to verify is another function is a ZO(ZO)Callable.
Multiple functions are defined in the submodule _ZOCallable.functions_, most of them are ZOZOCallables, all of them are ZOCallables. 
"""
from ZOCallable.ZOCallable import ZOCallable, verify_ZOCallable, normalize_ZOCallable, vectorize_ZOCallable
from ZOCallable import functions


"""The module ZOCallable the defintion of the ZOCallable class, used to type functions. ZO stands for Zero-One."""
from typing import Callable
import inspect
import numpy as np

class _ZOCMetaclass(type):
    """
    The Metaclass used to define a ZOZOCallable.
    Things are done like this to allow the check isinstance(func, ZOCallable) to return
    True if the function satisfies the conditions, even if it is not an instance of the class
    but a function or a lambda function.
    """

    rounding: int = 5
    test_vectorization: bool = False,
    test_values = np.linspace(0 + 1/101, 1 - 1/101, 99)

    def __instancecheck__(cls, func) -> bool:
        if not (
            isinstance(func, Callable) #The func must be a callable
            and len(inspect.signature(func).parameters.keys()) == 1 # the func must have one parameter
        ):
            return False
        f0 = func(0.)
        f1 = func(1.)
        if isinstance(f0, np.ndarray): # if the output is an array, the function is vectorized.
            try:
                f0 = float(f0)
                f1 = float(f1) # we assume f1 is also a np.ndarray in this case
            except ValueError:
                return False # We need to be sure the output is a float
         # The function must satisfy f(0) = 0 and f(1) = 1, it checkes also if the output is a number.
        if not (round(f0, _ZOCMetaclass.rounding) == 0 and round(f1, _ZOCMetaclass.rounding) == 1):
            return False
        if cls is ZOCallable: # No further check is the asked class is ZOCallable
            return True
        if _ZOCMetaclass.test_vectorization: # If we test the vectorization, we verify the output is an array of float
            outputs = func(_ZOCMetaclass.test_values)
            if not isinstance(outputs, np.ndarray) or not isinstance(outputs[0], float):
                return False
            # And we verify all these floats are in [0, 1]
            return np.all(0 <= np.round(outputs, _ZOCMetaclass.rounding) <= 1)
        else:
            # If we don't care about vectorization, we still test the range of the function.
            return all(0 <= round(float(func(t)), _ZOCMetaclass.rounding) <= 1 for t in _ZOCMetaclass.test_values)

class ZOCallable(metaclass=_ZOCMetaclass):
    """
    The ZOCallable isn't meant to be instanciated.
    This class is only a type hint for functions f with
    f(0) = 0, f(1) = 1 and f : [0, 1] -> R.
    """

    def __call__(self, x: float) -> float: ...

class ZOZOCallable(ZOCallable, metaclass=_ZOCMetaclass):
    """
    The ZOZOCallable isn't meant to be instanciated.
    This class is only a type hint for functions f with
    f(0) = 0, f(1) = 1 and f : [0, 1] -> [0, 1].
    """

    def __call__(self, x: float) -> float: ...

def verify_ZOCallable(ZOC, rounding: int=5):
    """
    Verify if the provided function is a ZOCallable.

    Params:
    ----
    - rounding: int = 5, the number of significative numbers that are kept to proceed to the float comparisons.
    This is used to allow the calculations errors of python.

    Returns:
    ---
    - is_ZOC: bool, whether the provided function is a ZOCallable.
    """
    _ZOCMetaclass.rounding = rounding
    return isinstance(ZOC, ZOCallable)

def normalize_ZOCallable(unnormalized_callable: Callable[[float], float]):
    """
    Normalize a function to be a ZOCallable.
    
    Params:
    ----
    - unnormalized_callable: Callable (float) -> float. A function to be normalized.

    Returns:
    ----
    - ZOC, a function (float) -> (float) satisfying ZOC(0) = 0 and ZOC(1) = 1.

    Raises:
    ----
    - ValueError if unnormalized_callable cannot be normalized, meaning unnormalized_callable(1) == 0.
    """
    u1 = unnormalized_callable(1)
    if u1 == 0:
        raise ValueError("This function cannot be normalized as a ZOCallable.")
    u0 = unnormalized_callable(0)
    return lambda t: (unnormalized_callable(t) - u0) / (u1 - u0)

def vectorize_ZOCallable(unvectorized_ZOC: Callable[[float], float]):
    """
    Vectorize a ZOCallable.

    In order to be used directly on a numpy array, most functions need to be vectorized.
    However, np.vectorize transforms the params from 'x' (or wathever it was) to (*args, **kwargs)
    while a ZOCallable need to have only one parameter.
    This function perform the vectorization with numpy and replaces the input parameter to be
    accepted as a vectorized ZOCallable.
    """
    vectorized_ZOC = np.vectorize(unvectorized_ZOC)
    return lambda x: vectorized_ZOC(x)

def verify_ZOZOCallable(ZOC, rounding: int=5, test_vectorizaiton: bool = False, points: int = 101):
    """
    Verify if the provided function is a ZOZOCallable.

    Params:
    ----
    - rounding: int = 5, the number of significative numbers that are kept to proceed to the float comparisons.
    This is used to allow the calculations errors of python.
    - test_vectorization: bool = False. If true, the function is also tested if it is vectorized.
    - points: int = 101, >= 3 the number of points used to test the assertion 0 <= f(x) <= 1.

    Returns:
    ---
    - is_ZOZOC: bool, whether the provided function is a ZOZOCallable.

    Raises:
    ----
    - ValueError: if points is <= 2.
    """
    if points <= 2:
        raise ValueError(f"{points} points are not enough.")
    _ZOCMetaclass.rounding = rounding
    _ZOCMetaclass.test_vectorization = test_vectorizaiton
    _ZOCMetaclass.test_values = np.linspace(0 + 1/points, 1 - 1/points, points - 2)
    return isinstance(ZOC, ZOZOCallable)



from setuptools import setup, find_packages

setup(
    name='ZOCallable',
    author="Tanguy Dugas du Villard",
    author_mail="tanguy.dugas01@gmail.com",
    version='1.0.1',
    description="ZOCallable is a library defining multiple functions f : [0, 1] -> R, and satisfying the condition f(0) = 0 and f(1) = 1. They can be used as transitions.",
    packages=find_packages(),
    install_requires=[
        'numpy',
    ],
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Tanguy-ddv/pygame-arts",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Mathematics"
    ],
    python_requires='>=3.6'
)


"""The functions submodule contains some ZOCs and ZOZOCs."""
import numpy as np
from .ZOCallable import ZOCallable, ZOZOCallable, normalize_ZOCallable, vectorize_ZOCallable

linear = lambda x:x

def power_in(n) -> ZOZOCallable:
    """Power_in functions are the polynomial functions f(x) = x^n."""
    return lambda x: np.pow(x, n)

def power_out(n) -> ZOZOCallable:
    """Power_out functions are the polynomial functions f(x) = 1 - (1-x)^n."""
    return lambda x: 1 - np.pow(1 - x, n)

def power_in_out(n) -> ZOZOCallable:
    """Power_out functions are a combination of a power_in for x < 0.5 and power_out for x > 0.5."""
    return vectorize_ZOCallable(lambda x: (2**(n-1)) *np.pow(x, n) if x < 0.5 else 1 - np.pow(1 - x, n)*(2**(n-1)))

square_in = power_in(2)
square_out = power_out(2)
square_in_out = power_in_out(2)
root_out = lambda x: np.sqrt(x)
root_in = lambda x: 1 - np.sqrt(1 - x)
    
def cubic_bezier(x1, y1, x2, y2, precision: float = 2**(-8)) -> ZOCallable:
    """
    Return a ZOCallable following a cubic bezier curve.
    
    Params:
    ----
    - x1, y1, x2, y2: float, The control points of the bezier curve, with the end and start points at (0, 0) and (1, 1).
    x1 and x2 must be in [0, 1], for the cubic_bezier to be a ZOZOCallable, y1 and y2 must be in [0, 1], otherwise the 
    returned function is only a ZOCallable
    The bezier curve is created following css notations, and can be explored here https://cubic-bezier.com
    - precision: float << 1, used to approximate (see warning below)

    Returns:
    ----
    - ease: ZOCallable, the function y(x) following the bezier curve.
    
    Examples:
    ---
    ```python
    ease = cubic_bezier(0.25, 0.1, 0.25, 1.)
    ease_in = cubic_bezier(0.42, 0., 1, 1.)
    ease_out = cubic_bezier(0., 0., 0.58, 1.)
    ease_in_out = cubic_bezier(0.42, 0., 0.58, 1.)
    ```
    these functions are already defined in the same module.
    

    Warning:
    ---
    As bezier curves are parametric (i.e. defined as (x(t), y(t)) with t in [0, 1]) and as there is no mathematical
    formula for y(x), it is approach via a dichotomy method, with a given precision. Thus, for a very smooth curve,
    it is adviced to a high precision. For most applications, 2^(-8) is precise enough. 
    """

    x0 = 0
    y0 = 0
    x3 = 1
    y3 = 1
    if x1 < 0 or x1 > 1 or x2 < 0 or x2 > 1:
        raise ValueError(f"x1 and x2 must be in [0, 1], got {x1} and {x2}")

    bezier_curve_x = lambda t: (1.-t)**3 * x0 + 3*(1.-t)**2*t * x1 + 3*(1.-t)*t**2 * x2 + t**3 * x3
    bezier_curve_y = lambda t: (1.-t)**3 * y0 + 3*(1.-t)**2*t * y1 + 3*(1.-t)*t**2 * y2 + t**3 * y3
    
    def find_y(x):
        if x < 0 or x > 1:
            raise ValueError(f"The input should be inside [0, 1], got {x}")
        if x == 0 or x == 1:
            return x
        t_left = 0.
        t_right = 1.
        x_left = bezier_curve_x(t_left)
        x_right = bezier_curve_x(t_right)
        while abs(x_right - x_left) > precision: # Dichotomy to find the t.
            x_center = bezier_curve_x((t_right + t_left)/2)
            if x_center > x:
                t_right = (t_right + t_left)/2
                x_right = x_center
            else:
                t_left = (t_right + t_left)/2
                x_left = x_center

        return bezier_curve_y((t_right + t_left)/2)
    
    return vectorize_ZOCallable(find_y)

ease = cubic_bezier(0.25, 0.1, 0.25, 1.)
ease_in = cubic_bezier(0.42, 0., 1, 1.)
ease_out = cubic_bezier(0., 0., 0.58, 1.)
ease_in_out = cubic_bezier(0.42, 0., 0.58, 1.)

def bounce_out(n: int) -> ZOZOCallable:
    """
    Return a ZOCallable that looks like bounces for large values.
    
    Params:
    ---
    - n: int >= 0, the number of bounces.

    Raises:
    ----
    ValueError, if n < 0.
    """
    if n < 0 or not isinstance(n, int):
        raise ValueError(f"{n} is not an acceptable argument for the number of bounce.")
    def bounce_n(x):
        if x == 0:  # Handle the edge case to avoid division by zero
            return x
        new_x = (n+1) * np.pi * np.power(x, 3/2)
        sinc = np.sin(new_x) / new_x
        return 1 - np.abs(sinc)
    bounce_n = vectorize_ZOCallable(bounce_n)
    return bounce_n

def bounce_in(n: int):
    """
    Return a ZOCallable that looks like bounces for small values.
    
    Params:
    ---
    - n: int >= 0, the number of bounces.

    Raises:
    ----
    ValueError, if n < 0.
    """

    bounce_in_n = bounce_out(n)
    return lambda x:1 - bounce_in_n(1 - x)

def jump(n: int) -> ZOZOCallable:
    """
    Return a ZOCallable being successive jumps.
    
    Params:
    ---
    - n: int >= 0, the number of jump.

    Returns:
    ----
    jump_n: a ZOZOCallable, whose graph looks like jumps, or stairs. The function as 1 jump at t = 1, the other are evenly spread between t = 0 and t = 1

    Raises:
    ----
    ValueError, if n <= 0.
    """

    if n <= 0:
        raise ValueError(f"{n} is not an acceptable argument for the number of jumps.")
    return lambda x: np.round(x*n)/n

exp_in = normalize_ZOCallable(np.exp)
exp_out = normalize_ZOCallable(lambda t: 1 - (1 - np.exp(t)))


# functions taken from https://github.com/semitable/easing-functions/blob/master/easing_functions/easing.py

sin_in = lambda x: np.sin((x - 1) * np.pi / 2) + 1
sin_out = lambda x: np.sin(x * np.pi / 2)
sin_in_out = lambda x: 0.5 * (1 - np.cos(x * np.pi))

circular_in = lambda x: 1 - root_in(1 - square_out(x))
circular_out = lambda x: 1 - root_out(1 - square_in(x))
circular_in_out = vectorize_ZOCallable(lambda x: 0.5 * (np.sqrt(-((2 * x) - 3) * ((2 * x) - 1)) + 1) if x > 0.5 else 0.5 * (1 - np.sqrt(1 - 4 * np.pow(x, 2))))

elastic_in = lambda x: np.sin(13 * np.pi / 2 * x)  / np.pow(2, 10 * (1 - x))
elastic_out = lambda x: np.sin(-13 * np.pi / 2 * (x + 1)) * np.pow(2, -10 * x) + 1
elastic_in_out = vectorize_ZOCallable(
    lambda x: (
        0.5 * np.sin(13 * np.pi / 2 * (2 * x)) * np.pow(2, 10 * ((2 * x) - 1))
    ) if x< 0.5 else (
        0.5 * ( np.sin(-13 * np.pi / 2 * ((2 * x - 1) + 1)) * np.pow(2, -10 * (2 * x - 1)) + 2)
    )
)

back_in = lambda x: np.pow(x, 3) - x * np.sin(x * np.pi)
back_out = lambda x: 1 - back_in(1 - x)
back_in_out = vectorize_ZOCallable(lambda x: 0.5*back_in(2*x) if x < 0.5 else 0.5*(1- back_in(1 - 2*x)))




from ZOCallable.ZOCallable import verify_ZOCallable, verify_ZOZOCallable, ZOZOCallable, ZOCallable
from typing import Callable
import unittest
class TestZOCallable(unittest.TestCase):

    def test_ZOCallable(self):
        func = lambda x:x
        self.assertTrue(verify_ZOCallable(func, 3), "lambda x:x should be a ZOCallable.")
        func = lambda x:x**2
        self.assertTrue(verify_ZOCallable(func, 3), "lambda x:x**2 should be a ZOCallable.")
        func = lambda x:2*x
        self.assertFalse(verify_ZOCallable(func, 3), "lambda x:2*x shouldn't be a ZOCallable as f(1) = 2.")
        func = lambda x:1-x
        self.assertFalse(verify_ZOCallable(func, 3), "lambda x:(1-x) shouldn't be a ZOCallable as f(0) = 1.")
        func = lambda x:2*x if x <0.75 else 6/4 - x/2 
        self.assertTrue(verify_ZOCallable(func, 3), "lambda x:2*x if x <0.75 else 6/4 - x/2 should be a ZOCallable even if it goes above 1.")
        func = lambda x:-x if x != 1 else 1
        self.assertTrue(verify_ZOCallable(func, 3), "x:-x if x != 1 else 1 should be a ZOCallable even if it goes below 0 and is not continous.")
        func = 1
        self.assertFalse(verify_ZOCallable(func, 3), "1 shouldn't be a ZOCallable as it is not a Callable.")
        func = lambda x,y: x + y
        self.assertFalse(verify_ZOCallable(func, 3), "lambda x,y: x + y shouldn't be a ZOCallable as it has 2 parameters.")

    def test_ZOZOCallable(self):
        func = lambda x:x
        self.assertTrue(verify_ZOZOCallable(func, 3), "lambda x:x should be a ZOZOCallable.")
        func = lambda x:x**2
        self.assertTrue(verify_ZOZOCallable(func, 3), "lambda x:x**2 should be a ZOZOCallable.")
        func = lambda x:2*x
        self.assertFalse(verify_ZOZOCallable(func, 3), "lambda x:2*x shouldn't be a ZOZOCallable as f(1) = 2.")
        func = lambda x:1-x
        self.assertFalse(verify_ZOZOCallable(func, 3), "lambda x:(1-x) shouldn't be a ZOZOCallable as f(0) = 1.")
        func = lambda x:2*x if x <0.75 else 6/4 - x/2 
        self.assertFalse(verify_ZOZOCallable(func, 3), "lambda x:2*x if x <0.75 else 6/4 - x/2 shouldn't be a ZOZOCallable as it goes above 1.")
        func = lambda x:-x if x != 1 else 1
        self.assertFalse(verify_ZOZOCallable(func, 3), "x:-x if x != 1 else 1 shouldn't be a ZOZOCallable even if it goes below 0.")
        func = 1
        self.assertFalse(verify_ZOZOCallable(func, 3), "1 shouldn't be a ZOZOCallable as it is not a Callable.")
        func = lambda x,y: x + y
        self.assertFalse(verify_ZOZOCallable(func, 3), "lambda x,y: x + y shouldn't be a ZOZOCallable as it has 2 parameters.")

    def test_functions(self):
        from ZOCallable.functions import linear, square_in, square_in_out, square_out, root_in, root_out
        funcs = [linear, square_in, square_out, square_in_out, root_in, root_out]
        for i, func in enumerate(funcs):
            self.assertTrue(verify_ZOCallable(func, 3), f"The {i+1}-th function of the basic list should be a ZOCallable.")
            self.assertTrue(verify_ZOZOCallable(func, 3), f"The {i+1}-th function of the basic list should be a ZOZOCallable.")
        from ZOCallable.functions import power_in, power_in_out, power_out
        funcs = sum([[power_in(i), power_out(i), power_in_out(i)] for i in range(3, 8)], [])
        for i, func in enumerate(funcs):
            self.assertTrue(verify_ZOCallable(func, 3), f"The power_{i%3}({i//3 + 3} function should be a ZOCallable.")
            self.assertTrue(verify_ZOZOCallable(func, 3), f"The power_{i%3}({i//3 + 3} function should be a ZOZOCallable.")
        from ZOCallable.functions import ease, ease_in, ease_in_out, ease_out
        funcs = [ease, ease_in, ease_in_out, ease_out]
        for i, func in enumerate(funcs):
            self.assertTrue(verify_ZOCallable(func, 3), f"The {i+1}-th function of the ease list should be a ZOCallable.")
            self.assertTrue(verify_ZOZOCallable(func, 3), f"The {i+1}-th function of the ease list should be a ZOZOCallable.")
        from ZOCallable.functions import jump, bounce_in, bounce_out
        funcs = sum([[jump(i), bounce_in(i), bounce_out(i)] for i in range(3, 8)], [])
        for i, func in enumerate(funcs):
            self.assertTrue(verify_ZOCallable(func, 3), f"The special{i%3}({i//3} function should be a ZOCallable.")
            self.assertTrue(verify_ZOZOCallable(func, 3), f"The special{i%3}({i//3} function should be a ZOZOCallable.")
        from ZOCallable.functions import exp_in, exp_out, circular_in, circular_in_out, circular_out, back_in, back_in_out, back_out, elastic_in, elastic_in_out, elastic_out
        funcs = [exp_in, exp_out, circular_in, circular_in_out, circular_out, back_in, back_in_out, back_out, elastic_in, elastic_in_out, elastic_out]
        for i, func in enumerate(funcs):
            self.assertTrue(verify_ZOCallable(func, 3), f"The {i+1}-th function of the extra list should be a ZOCallable.")
            if i < 5: # the back ad elastic functions are supposed to not be ZOZOC
                self.assertTrue(verify_ZOZOCallable(func, 3), f"The {i+1}-th function of the extra list should be a ZOZOCallable.")
        
    def test_plots(self):
        from ZOCallable.functions import circular_in, circular_in_out, circular_out, elastic_in, elastic_in_out, elastic_out, ease, ease_in, ease_in_out, ease_out
        from ZOCallable.plot import plot_ZOCallable, plot_ZOZOCallable
        plot_ZOZOCallable(circular_in, circular_in_out, circular_out, vectorized=True, save_path="ZOZOC.png", labels=["Circular_in", "Circular_in_out", "Circular_out"])
        plot_ZOCallable(elastic_in, elastic_in_out, elastic_out, vectorized=True, save_path="ZOC.png", labels=["Elastic_in", "Elastic_in_out", "Elastic_out"])
        plot_ZOCallable(ease, ease_in, ease_in_out, ease_out, vectorized=True, save_path="ease.png", labels=["ease", "ease_in", "ease_in_out", "ease_out"])

    def test_metaclass(self):
        self.assertTrue(issubclass(ZOZOCallable, ZOCallable), "ZOZOCallable should be a subclass of ZOCallble.")
        self.assertFalse(issubclass(ZOCallable, ZOZOCallable), "ZOCallable shouldn't be a subclass of ZOZOCallable.")
        self.assertTrue(issubclass(ZOCallable, Callable), "ZOallable should be a subclass of Callable.")
        self.assertTrue(issubclass(ZOZOCallable, Callable), "ZOZOCallable should be a subclass of ZOCallble.")
        self.assertFalse(issubclass(int, ZOCallable), "int shouldn't be a subclass of ZOCallable.")
    